﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using OpenTK.Graphics.OpenGL;

namespace HW09CPC
{
    /// Helper-Klasse um custom Mipmaps zu erstellen (speziell das Maximum zu berechnen)
    class MipmapGenerator
    {
        private Form1 mParent;

        private GLTextureRgba2D mSourceTex;
        private GLProgram mProgram;

        private List<GLFbo2D> mMipmapFbos;
        private List<GLFbo2D> mTempFbos;
        private GLTextureRgba2D mMipmapTexture;
        private GLTextureRgba2D mTempTexture;

        private GLTextureRgba2D lastWriteTex;
        private int mLastLevel;

        /// Konstruktor
        /// <param name="_parent">das Parent-Form1</param>
        /// <param name="_baseTex">die Basis-Textur</param>
        /// <param name="_program">das zu verwendente GL-Program</param>
        public MipmapGenerator(Form1 _parent, GLTextureRgba2D _baseTex, GLProgram _program)
        {
            mParent = _parent;
            mSourceTex = _baseTex;
            mProgram = _program;
            
            mMipmapFbos = null;
            mTempFbos = null;

            mMipmapTexture = null;
            mTempTexture = null;

            setup();
        }

        /// Skaliert den Mipmap-Generator
        public void resize()
        {
            setup();
        }

        private void setup()
        {
            // cleanup

            if (mMipmapFbos != null)
            {
                foreach (GLFbo2D fbo in mMipmapFbos)
                    fbo.delete();

                mMipmapFbos.Clear();
            }

            if (mTempFbos != null)
            {
                foreach (GLFbo2D fbo in mTempFbos)
                    fbo.delete();

                mTempFbos.Clear();
            }

            if (mMipmapTexture != null)
            {
                mMipmapTexture.deleteTexture();
            }

            if (mTempTexture != null)
            {
                mTempTexture.deleteTexture();
            }

            // reinit
                
            mMipmapFbos = new List<GLFbo2D>();
            mTempFbos = new List<GLFbo2D>();

            lastWriteTex = null;
            mLastLevel = 0;

            initLevels();
        }
        
        private void initLevels()
        {
            // calc new power of two dimentions

            uint npot;

            if(mSourceTex.getWidth() > mSourceTex.getHeight())
                npot = getNextPowerOfTwo(mSourceTex.getWidth());
            else
                npot = getNextPowerOfTwo(mSourceTex.getHeight());

            // init the mipmapped texture and all mipmap level fbos

            int level = 0;

            mMipmapTexture = new GLTextureRgba2D("mipmap" + level, npot, npot, OpenTK.Graphics.OpenGL.TextureUnit.Texture2, true, false, true, null);
            mMipmapTexture.setNearestFiltering();
            mMipmapTexture.setWrapMode(All.ClampToEdge, All.ClampToEdge, All.ClampToEdge);
            mMipmapFbos.Add(new GLFbo2D(mMipmapTexture));

            mTempTexture = new GLTextureRgba2D("temp" + level, npot, npot, OpenTK.Graphics.OpenGL.TextureUnit.Texture3, true, false, true, null);
            mTempTexture.setNearestFiltering();
            mTempTexture.setWrapMode(All.ClampToEdge, All.ClampToEdge, All.ClampToEdge);
            mTempFbos.Add(new GLFbo2D(mTempTexture));
            
            for(uint size = npot/2; size > 0; size /= 2) // change to a better for loop
            {
                level++;
                mMipmapFbos.Add(new GLFbo2D(mMipmapTexture, level));
                mTempFbos.Add(new GLFbo2D(mTempTexture, level));
            }
        }

        private void initBase()
        {
            // render source texture into base mipmap level

            GLFbo2D baseLevel = mMipmapFbos.ElementAt<GLFbo2D>(0);

            baseLevel.bindClear();

            // only draw into region the size of the source texture
            GL.Viewport(0, 0, (int)mSourceTex.getWidth(), (int)mSourceTex.getHeight());

            mParent.drawScreenQuad(mSourceTex);

            baseLevel.unbind();
        }

        /// Berechnet die custom Mipmaps (speziell das Maximum)
        public void calcMipmaps()
        {
            GL.Disable(EnableCap.Blend); // without this it is not working ... strange?!

            initBase();

            mProgram.bind();

            // Quad transform
            int stULoc = mProgram.GetUniformLocation("scaleTrans");
            GL.Uniform4(stULoc, 1.0f, 1.0f, 0.0f, 0.0f);

            GLTextureRgba2D readTex;
            GLFbo2D writeFbo;

            // PING PONG
            
            int writeLevel = 0;

            GLFbo2D lf = null;

            for (int readLevel = 0; readLevel < mMipmapFbos.Count - 1; readLevel++)
            {
                writeLevel = readLevel + 1;

                if (readLevel % 2 == 0)
                {                
                    readTex = mMipmapFbos.ElementAt<GLFbo2D>(readLevel).getColorAttachment();
                    writeFbo = mTempFbos.ElementAt<GLFbo2D>(writeLevel);
                }
                else
                {
                    readTex = mTempFbos.ElementAt<GLFbo2D>(readLevel).getColorAttachment();
                    writeFbo = mMipmapFbos.ElementAt<GLFbo2D>(writeLevel);
                }
                
                writeFbo.getColorAttachment().bind();
                
                GL.TexParameter(TextureTarget.Texture2D, TextureParameterName.TextureBaseLevel, readLevel);
                GL.TexParameter(TextureTarget.Texture2D, TextureParameterName.TextureMaxLevel, readLevel);
                writeFbo.getColorAttachment().unbind();

                // Quad texture
                readTex.bind();
                int readTexULoc = mProgram.GetUniformLocation("readTex");
                GL.Uniform1(readTexULoc, readTex.getTextureUnit() - TextureUnit.Texture0);

                int readLevelULoc = mProgram.GetUniformLocation("readLevel");
                GL.Uniform1(readLevelULoc, (float)readLevel);
                          
                int w = (int)(writeFbo.getColorAttachment().getWidth()/Math.Pow(2, writeLevel));
                int h = (int)(writeFbo.getColorAttachment().getHeight()/Math.Pow(2, writeLevel));

                float hts = 0.25f / (float)w;
                int halfTexelSizeULoc = mProgram.GetUniformLocation("halfTexelSize");
                GL.Uniform1(halfTexelSizeULoc, hts);

                writeFbo.bindClear();

                GL.Viewport(0, 0, w, h);

                mParent.drawScreenQuadGeometry();
                writeFbo.unbind();

                readTex.unbind();

                writeFbo.getColorAttachment().bind();

                GL.TexParameter(TextureTarget.Texture2D, TextureParameterName.TextureBaseLevel, 0);
                GL.TexParameter(TextureTarget.Texture2D, TextureParameterName.TextureMaxLevel, mMipmapFbos.Count - 1);
                writeFbo.getColorAttachment().unbind();

                lf = writeFbo;
                lastWriteTex = writeFbo.getColorAttachment();
            }

            // just lazy vars
            mLastLevel = writeLevel;
            mProgram.unbind();

            GL.Enable(EnableCap.Blend);
        }

        /// Liefert die Mipmap-Textur zurück
        /// <returns>die Mipmap-Textur</returns>
        public GLTextureRgba2D getMipmapTex()
        {
            return mMipmapTexture;
        }

        /// Liefert die Temp-Textur (Helper) zurück
        /// <returns>die Temp-Textur</returns>
        public GLTextureRgba2D getTempTex()
        {
            return mTempTexture;
        }

        /// Liefert die Textur zurück welche das Maximum enhält
        /// <returns>die Maximum-Textur</returns>
        public GLTextureRgba2D getMaxTex()
        {
            return lastWriteTex;
        }

        /// Liefert das maximale Level zurück
        /// <returns>das maximale Level</returns>
        public int getMaxLevel()
        {
            return mLastLevel;
        }

        private uint getNextPowerOfTwo(uint _value)
        {
            uint result = 2;

            while (result < _value)
            {
                result *= 2;
            }

            return result;
        }
    }
}
