﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using OpenTK.Graphics.OpenGL;

namespace HW09CPC
{
    /// Kapselt ein FBO mit 2D Color-Attachment
    class GLFbo2D
    {
        protected int mFboHandle;
	    protected FramebufferTarget mTarget;
        protected DrawBufferMode mDrawBuffer;
	    protected ReadBufferMode mReadBuffer;

        protected GLTextureRgba2D mColorAttachment;
        protected int[] mBeforeBindViewport;

        /// Konstruktor
        /// <param name="_colorAttachment">das Color-Attachment</param>
        /// <param name="_level">das Level des Color-Attachments welches an das FBO gebunden werden soll</param>
        public GLFbo2D(GLTextureRgba2D _colorAttachment, int _level = 0)
        {
            mColorAttachment = _colorAttachment;
	        initMembers();

            build(_level);
        }

        /// Löscht das FBO
        public void delete()
        {
            deleteHandle();
        }

        private void initMembers()
        {
	        mFboHandle = 0;
	        mTarget = FramebufferTarget.Framebuffer;
            mDrawBuffer = DrawBufferMode.ColorAttachment0;
	        mReadBuffer = ReadBufferMode.ColorAttachment0;
        }

        protected void generateHandle()
        {
	        if(mFboHandle > 0)
		        return;

	        GL.GenFramebuffers(1, out mFboHandle);
        }

        protected void deleteHandle()
        {
	        if(mFboHandle > 0) GL.DeleteFramebuffers(1, ref mFboHandle);
        }

        /// Legt das FBO an.
        /// <param name="_level">das Level des Color-Attachments welches an das FBO gebunden werden soll</param>
        public void build(int _level = 0)
        {
	        generateHandle();
            bind();

            GL.FramebufferTexture2D(FramebufferTarget.Framebuffer, FramebufferAttachment.ColorAttachment0, TextureTarget.Texture2D, mColorAttachment.getHandle(), _level);
          
	        Utility.GLErrorCheck("GLFbo2D::build");
	        getFboError();

            unbind();
        }

        protected void setViewport()
        {
	        GL.Viewport(0, 0, (int)mColorAttachment.getWidth(), (int)mColorAttachment.getHeight());
        }

        protected void resetViewport()
        {
	        GL.Viewport(mBeforeBindViewport[0], mBeforeBindViewport[1], mBeforeBindViewport[2], mBeforeBindViewport[3]); 
        }

        protected void backupViewport()
        {
            mBeforeBindViewport = new int[4];
	        GL.GetInteger(GetPName.Viewport, mBeforeBindViewport);
        }

        /// Aktiviert das FBO
        public void bind()
        {	
	        backupViewport();

	        GL.BindFramebuffer(mTarget, mFboHandle);
            GL.DrawBuffer(mDrawBuffer);
	        GL.ReadBuffer(mReadBuffer);

	        setViewport();
        }

        /// Deaktiviert das FBO
        public void unbind()
        {
	        GL.BindFramebuffer(mTarget, 0);
	        GL.DrawBuffer(DrawBufferMode.Back);
	        GL.ReadBuffer(ReadBufferMode.Back);

	        resetViewport();
        }

        /// Aktiviert das FBO und cleared seinen Buffer
        public void bindClear()
        {
	        bind();
	        clear();
        }

        /// Cleared den FBO Buffer
        public void clear()
        {
	        GL.Clear(ClearBufferMask.ColorBufferBit);
        }
        
        /// Skaliert das FBO
        /// <param name="_width">die neue Breite</param>
        /// <param name="_height">die neue Höhe</param>
        public void resize(uint _width, uint _height)
        {
	        mColorAttachment.resize(_width, _height);
        }

        /// Berechnet die Mipmap-Stufen neu
        public void updateMipmaps()
        {
	        mColorAttachment.updateMipmap();
        }

        /// Generiert die Mipmap-Stufen
        public void generateMipmaps()
        {
	        mColorAttachment.generateMipmap();
        }

        protected uint getWidth()
        {
            return mColorAttachment.getWidth();
        }

        protected uint getHeight()
        {
            return mColorAttachment.getHeight();
        }

        /// Liefert das Color-Attachment zurück
        /// <returns>das Color-Attachment</returns>
        public GLTextureRgba2D getColorAttachment()
        {
            return mColorAttachment;
        }

        /// Überpüft auf FBO GL-Error
        public void getFboError()
        {
            FramebufferErrorCode status = GL.CheckFramebufferStatus(FramebufferTarget.Framebuffer);

            if (status != FramebufferErrorCode.FramebufferComplete)
	        {
		        string error = "unknown error";

		        switch (status)
		        {
                    case FramebufferErrorCode.FramebufferUndefined:
				        error = "GL_FRAMEBUFFER_UNDEFINED";
				        break;
                    case FramebufferErrorCode.FramebufferIncompleteAttachment:
				        error = "GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT";
				        break;
                    case FramebufferErrorCode.FramebufferIncompleteMissingAttachment:
				        error = "GL_FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT";
				        break;
                    case FramebufferErrorCode.FramebufferIncompleteDrawBuffer:
				        error = "GL_FRAMEBUFFER_INCOMPLETE_DRAW_BUFFER";
				        break;
                    case FramebufferErrorCode.FramebufferIncompleteReadBuffer:
				        error = "GL_FRAMEBUFFER_INCOMPLETE_READ_BUFFER";
				        break;
                    case FramebufferErrorCode.FramebufferUnsupported:
				        error = "GL_FRAMEBUFFER_UNSUPPORTED";
				        break;
                    case FramebufferErrorCode.FramebufferIncompleteMultisample:
				        error = "GL_FRAMEBUFFER_INCOMPLETE_MULTISAMPLE";
				        break;
                    case FramebufferErrorCode.FramebufferIncompleteLayerTargets:
				        error = "GL_FRAMEBUFFER_INCOMPLETE_LAYER_TARGETS";
				        break;
			        default:
				        break;
		        }

                System.Console.WriteLine("FBO Error: " + error);
		        throw new Exception("FBO Error");
	        }
	        else
	        {
                //System.Console.WriteLine("FBO complete");
	        }
        }
    }
}
