﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using OpenTK.Graphics.OpenGL;

namespace HW09CPC
{
    /// Stellt die Basis-Klasse der GLTexture Kapselungen dar.
    public abstract class GLTexture
    {
        protected bool mIsLoaded;

        protected float[] mData;
        protected uint mWidth, mHeight;

        protected int mHandle;
        protected TextureUnit mTextureUnit;
        protected TextureTarget mTarget;

        protected PixelInternalFormat mInternalFormat;
        protected PixelFormat mFormat;
        protected PixelType mType;

        protected TextureMinFilter mMinFilter;
        protected TextureMagFilter mMagFilter;
        protected All mWrapS;
        protected All mWrapT;
        protected All mWrapR;

        protected string mName;
        protected bool mIsMipmapped;

        protected GLTexture(string _name, uint _width, uint _height, TextureUnit _textureUnit, bool _generateMipmap, float[] _data)
        {
	        initMembers(_name, _width, _height, _textureUnit, _generateMipmap);
            
            if (_data == null)
                return;

	        mData = new float[_data.Length];
            for (uint p = 0; p < _data.Length; p++)
	        {
		        mData[p] = _data[p];
	        }
        }
        
        private void initMembers(string _name, uint _width, uint _height, TextureUnit _textureUnit, bool _generateMipmap)
        {
	        mIsLoaded = false;

	        mHandle = 0;
            mData = null;

            mWidth = _width;
	        mHeight = _height;	
	        mTextureUnit = _textureUnit;

	        mTarget = TextureTarget.Texture2D;

	        mInternalFormat = PixelInternalFormat.Rgba32f;
	        mFormat = PixelFormat.Rgba;
	        mType = PixelType.Float;

	        mMinFilter = TextureMinFilter.Linear;
	        mMagFilter = TextureMagFilter.Linear;
	        mWrapS = All.ClampToEdge;
            mWrapT = All.ClampToEdge;
            mWrapR = All.ClampToEdge;
	
	        mName = _name;		        
	        mIsMipmapped = _generateMipmap;
        }

        protected void generateHandle()
        {
	        if(mHandle > 0)
		        return;

	        GL.GenTextures(1, out mHandle);
        }

        /// Löscht die Textur
        public void deleteTexture()
        {
	        if(mHandle > 0)
	        {
		        GL.DeleteTextures(1, ref mHandle);
		        mHandle = 0;
	        }
        }

        /// Bindet die Textur
        public void bind()
        {
	        GL.ActiveTexture(mTextureUnit);      
            GL.BindTexture(mTarget, mHandle);
        }

        /// Bindet die Textur and die angegebene Textur-Einheit
        /// <param name="_textureUnit">die gewünschte Textur-Einheit</param>
        public void bind(TextureUnit _textureUnit)
        {
	        GL.ActiveTexture(_textureUnit);      
            GL.BindTexture(mTarget, mHandle);
	
	        //setParameter();
        }

        /// Löst die Textur
        public void unbind()
        {	     
            GL.BindTexture(mTarget, 0);
	        GL.ActiveTexture(TextureUnit.Texture0);	
        }

        /// Läd die Textur
        public void loadTexture()
        {
	        if(isLoaded())
		        return;

	        reloadTexture();
        }

        private void reloadTexture()
        {
	        generateHandle();
	        bind();
	        setParameter();
	        uploadTexture();
	        unbind();
	        mIsLoaded = true;
        }

        /// Läd die Textur in den Grafikspeicher und das angegebene Mipmap-Level
        /// <param name="_level">das gewünschte Mipmap-Level</param>
        public abstract void uploadTexture(int _level = 0);

        /// Liefert den Lade-Status der Textur zurück
        /// <returns>der Lade-Status</returns>
        public bool isLoaded()
        {
	        return mIsLoaded;
        }

        /// Liefert den Handle der Textur zurück
        /// <returns>der Handle</returns>
        public int getHandle()
        {
	        return mHandle;
        }

        /// Liefert die Textur-Einheit der Textur zurück 
        /// <returns>die Textur-Einheit</returns>
        public TextureUnit getTextureUnit()
        {
	        return mTextureUnit;
        }

        /// Setzt die Textur-Einheit 
        /// <param name="_textureUnit">die gewünschte Textur-Einheit</param>
        public void setTextureUnit(TextureUnit _textureUnit)
        {
	        mTextureUnit =_textureUnit;
        }

        /// Liefert die Breite der Textur zurück
        /// <returns>die Breite</returns>
        public uint getWidth()
        {
	        return mWidth;
        }

        /// Liefert die Höhe der Textur zurück
        /// <returns>die Höhe</returns>
        public uint getHeight()
        {
	        return mHeight;
        }

        /// Liefert den Mipmap-Status der Textur zurück
        /// <returns>der Mipmap-Status</returns>
        public bool isMipmapped()
        {
	        return mIsMipmapped;
        }

        /// Skaliert die Textur
        /// <param name="_width">die neue Breite</param>
        /// <param name="_height">die neue Höhe</param>
        public void resize(uint _width, uint _height)
        {
	        mWidth = _width;
	        mHeight = _height;

	        reloadTexture();
        }

        /// Generiert die Mipmap-Stufen
        public void generateMipmap()
        {
	        mIsMipmapped = true;

	        bind();
            
            setNearest();
            setParameter();

	        GL.GenerateMipmap(GenerateMipmapTarget.Texture2D);
	
	        unbind();
        }

        /// Berechnet die Mipmap-Stufen neu
        public void updateMipmap()
        {
	        if(isMipmapped())
		        generateMipmap();
        }

        /// Deaktiviert Mipmapping
        public void disableMipmapping()
        {
	        mIsMipmapped = false;

	        // loadData();
        }

        /// Setzt Nearest-Filterung
        public void setNearestFiltering()
        {
	        bind();
	        setNearest();
	        setParameter();
	        unbind();
        }

        /// Setzt Linear-Filterung
        public void setLinearFiltering()
        {
	        bind();
	        setLinear();
	        setParameter();
	        unbind();
        }

        protected void setNearest()
        {
            mMagFilter = TextureMagFilter.Nearest;
	
	        if(mIsMipmapped)
	        {
                mMinFilter = TextureMinFilter.NearestMipmapNearest; // maybe also support GL_LINEAR_MIPMAP_LINEAR	
	        }
	        else
	        {
                mMinFilter = TextureMinFilter.Nearest;
	        }
        }

        protected void setLinear()
        {
            mMagFilter = TextureMagFilter.Linear;

	        if(mIsMipmapped)
	        {
                if (mFormat == PixelFormat.DepthComponent)
                    mMinFilter = TextureMinFilter.LinearMipmapNearest; // stupid bug: if depth texture is displayed for debugging then mipmapping works for shadow (stupid idea but still), if not: mipmaps do not seem to get updated
		        else
                    mMinFilter = TextureMinFilter.LinearMipmapLinear; // maybe also support GL_LINEAR_MIPMAP_NEAREST
	        }
	        else
	        {
                mMinFilter = TextureMinFilter.Linear;
	        }
        }

        /// Setzt den Wrap-Modus
        /// <param name="_s">der S Wrap-Modus</param>
        /// <param name="_t">der T Wrap-Modus</param>
        /// <param name="_r">der R Wrap-Modus</param>
        public void setWrapMode(All _s, All _t, All _r)
        {
	        mWrapS = _s;
	        mWrapT = _t;
	        mWrapR = _r;

	        bind();
	        setParameter();
	        unbind();	
        }

        protected void setParameter()
        {
	        GL.TexParameter(mTarget, TextureParameterName.TextureWrapS, (int)mWrapS);
	        GL.TexParameter(mTarget, TextureParameterName.TextureWrapT, (int)mWrapT);
            GL.TexParameter(mTarget, TextureParameterName.TextureWrapR, (int)mWrapR);

            GL.TexParameter(mTarget, TextureParameterName.TextureMagFilter, (int)mMagFilter);
            GL.TexParameter(mTarget, TextureParameterName.TextureMinFilter, (int)mMinFilter);
        }
    }
}