﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using OpenTK.Graphics.OpenGL;

namespace HW09CPC
{
    /// Diese Klasse stellt die Geometrie-Daten dar und kapselt VAO und VBO
    class PrimitiveData
    { 
        private float[] mPrimitiveData; // Maybe let the VBO hold the primitveData, and reimplement the loding logic 
	    private GLVbo mVertexVbo;
	    private GLVao mVao;

	    private BufferUsageHint mVertexVboUsage;
        private PrimitiveDataInfo mPrimitiveDataInfo;

        /// Konstruktor
        /// <param name="_primitiveData">die Primitiv-Daten (für das VBO)</param>
        /// <param name="_primitiveDataInfo">die Attribut Beschreibung</param>
        /// <param name="_vertexVboUsage">die VBO-Usage</param>
        /// <param name="_loadVao">true: das VAO/VBO wird sofort geladen</param>
        public PrimitiveData(float[] _primitiveData, PrimitiveDataInfo _primitiveDataInfo, BufferUsageHint _vertexVboUsage, bool _loadVao)
        {
            mPrimitiveData = new float[_primitiveData.Length];

	        for(uint i = 0; i < _primitiveData.Length; i++)
	        {
		        mPrimitiveData[i] = _primitiveData[i];
	        }
	
	        mPrimitiveDataInfo = _primitiveDataInfo;
	        mVertexVboUsage = _vertexVboUsage;
	
	        mVao = null;
	        mVertexVbo = null;
            
            loadVertexVbo(mVertexVboUsage);
	
	        if(_loadVao)
		        loadVao();
        }

        /// Konstruktor
        /// <param name="_primitiveData">ein bereits erstelltes VBO</param>
        /// <param name="_primitiveDataInfo">die Attribut Beschreibung</param>
        /// <param name="_loadVao">true: das VAO/VBO wird sofort geladen</param>
        public PrimitiveData(GLVbo _vbo, PrimitiveDataInfo _primitiveDataInfo, bool _loadVao)
        {
            mPrimitiveDataInfo = _primitiveDataInfo;
            mVertexVboUsage = _vbo.getUsage();

            mVao = null;
            mVertexVbo = _vbo;

            if (_loadVao)
                loadVao();
        }
        
        private void loadVertexVbo(BufferUsageHint _vertexVboUsage)
        {
            if (mVertexVbo != null)
                mVertexVbo.loadVertexData(mPrimitiveData, _vertexVboUsage);
            else
                mVertexVbo = new GLVbo(mPrimitiveData, _vertexVboUsage);
        }

        /// liefert das Vertex-VBO zurück
        public GLVbo getVertexVbo()
        {
            return mVertexVbo;
        }

        /// Gibt VAO und VBO frei.
        /// <param name="_vbo">false: das VBO wird nicht gelöscht</param>
        public void delete(bool _vbo = true)
        {
            if (_vbo)
            {
                mVertexVbo.delete();
                mVertexVbo = null;
            }
            mVao.delete();
            mVao = null;
        }

        /// Aktiviert das VAO
        public void activate()
        {
	        mVao.bind();
        }

        /// Deaktiviert das VAO
        public void disable()
        {
	        mVao.unbind();
        }

        /// Zeichnet die Geometrie
        /// <param name="_first">der Index des Start-Elements</param>
        public void draw(int _first = 0)
        {
            draw(mPrimitiveDataInfo.getPrimitiveMode(), _first, mVertexVbo.getDataLength() / mPrimitiveDataInfo.getTotalOffset());
        }

        /// Zeichnet die Geometrie
        /// <param name="_first">der Index des Start-Primitivs</param>
        /// <param name="_count">die Anzahl der zu zeichneten Primitive</param>
        public void draw(int _first, int _count)
        {
            draw(mPrimitiveDataInfo.getPrimitiveMode(), _first, _count);
        }

        /// Zeichnet die Geometrie 
        /// <param name="_beginMode">die Primitiv-Art</param>
        public void draw(BeginMode _beginMode)
        {
            draw(_beginMode, 0, mPrimitiveData.Length / mPrimitiveDataInfo.getTotalOffset());
        }

        /// Zeichnet die Geometrie 
        /// <param name="_beginMode">die Primitiv-Art</param>
        /// <param name="_first">der Index des Start-Primitivs</param>
        /// <param name="_count">die Anzahl der zu zeichneten Primitive</param>
        public void draw(BeginMode _beginMode, int _first, int _count)
        {
            GL.DrawArrays(_beginMode, _first, _count);
        }

        private void loadVao()
        {	
	        activateVao();	
	        activateVertexVbo();

	        disableVao();
	        disableVertexVbo();
        }

        private void activateVao()
        {
	        // maybe put in own method
	        if(mVao == null)
		        mVao = new GLVao();

	        mVao.bind();
        }

        private void disableVao()
        {
	        mVao.unbind();
        }

        private void activateVertexVbo()
        {
            mVertexVbo.bind();

            for (int i = 0; i < mPrimitiveDataInfo.getAttributeCount(); i++)
            {
                GL.EnableVertexAttribArray(i);
                GL.VertexAttribPointer(i, mPrimitiveDataInfo.getLengthByIndex(i), VertexAttribPointerType.Float, false, mPrimitiveDataInfo.getTotalOffset() * sizeof(float), mPrimitiveDataInfo.getOffsetByIndex(i) * sizeof(float));
            }
        }

        private void activateVertexVbo(PrimitiveDataInfo _primitiveDataInfo)
        {	
	        mVertexVbo.bind();

            for (int i = 0; i < _primitiveDataInfo.getAttributeCount(); i++)
            {
                GL.EnableVertexAttribArray(i);
                GL.VertexAttribPointer(i, _primitiveDataInfo.getLengthByIndex(i), VertexAttribPointerType.Float, false, _primitiveDataInfo.getTotalOffset() * sizeof(float), _primitiveDataInfo.getOffsetByIndex(i) * sizeof(float));
            }
        }

        private void disableVertexVbo()
        {
            for (int i = 0; i < mPrimitiveDataInfo.getAttributeCount(); i++)
                GL.DisableVertexAttribArray(i);

	        mVertexVbo.unbind();
        }
    }
}
